from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from oarepo_model_builder.utils.verbose import log

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder
    from oarepo_model_builder.datatypes.datatypes import DataType


class OutputBuilder:
    TYPE = None

    def __init__(self, builder: ModelBuilder):
        self.builder = builder
        self.silent_exceptions = False

    def begin(self, current_model: DataType, schema):
        self.schema = schema
        self.current_model = current_model
        self.settings = schema["settings"]
        log.enter(2, "Creating %s", self.TYPE)
        self.silent_exceptions = False

    def finish(self):
        log.leave()

    def build(self, current_model, schema):
        self.begin(current_model, schema)
        self._build_node_internal(self.current_model)
        self.finish()

    def _build_node_internal(self, node):
        try:
            self.build_node(node)
        except Exception as e:
            if not self.silent_exceptions:
                self.silent_exceptions = True
                print(
                    f"Error on handling path {node.path}: {e}",
                    file=sys.stderr,
                )
            raise

    def build_children(self, parent_node):
        for child in parent_node.children.items():
            self._build_node_internal(child)

    def build_node(self, datatype: DataType):
        "Intended to be overriden in children"


TEMPLATES = {
    "setup_py": "templates/setup.py.jinja2",
}

__all__ = [
    "OutputBuilder",
    "TEMPLATES",
]



from oarepo_model_builder.datatypes import DataTypeComponent, datatypes


class RegularSortComponent(DataTypeComponent):
    eligible_datatypes = []

    def after_model_prepare(self, datatype, *, context, **kwargs):
        sort_options = []
        for node in datatype.deep_iter():
            sort_data = datatypes.call_components(
                node, "create_sort_options", context=context
            )
            if sort_data[0]:
                sort_options.extend(sort_data)
        datatype.definition["sortable"] = sort_options



import dataclasses
import keyword
import os
import re
from typing import Optional


@dataclasses.dataclass
class Import:
    import_path: str
    alias: Optional[str] = None

    @staticmethod
    def from_config(d):
        if isinstance(d, dict):
            return Import(d["import"], d.get("alias"))
        elif isinstance(d, (tuple, list)):
            return [Import.from_config(x) for x in d]

    def __hash__(self):
        return hash(self.import_path) ^ hash(self.alias)

    def __eq__(self, o):
        return self.import_path == o.import_path and self.alias == o.alias


def convert_name_to_python(name):
    if not name:
        return name

    # special case :)
    name = name.replace("@v", "_version")
    if name == "id":
        name = name.replace("id", "_id")
    # Replace any spaces or special characters in the string with an underscore
    identifier = re.sub(r"[^\w\s]", "_", name)
    identifier = re.sub(r"\s+", "_", identifier)

    # Ensure that the first character of the identifier is a letter or an underscore
    if identifier[0].isdigit():
        identifier = "_" + identifier

    # Ensure that the resulting identifier is not a reserved keyword in Python
    if keyword.iskeyword(identifier):
        identifier = identifier + "_"

    return identifier


def capitalize(s):
    return s[0].capitalize() + s[1:]


def convert_name_to_python_class(name):
    class_name = "".join([capitalize(word) for word in name.split("-")])
    class_name = "".join([capitalize(word) for word in class_name.split("_")])

    # Replace any spaces or special characters in the string with an underscore
    class_name = re.sub(r"[^\w\s]", "", class_name)
    class_name = re.sub(r"\s+", "_", class_name)

    if class_name[0].isdigit():
        class_name = "_" + class_name

    return class_name


def package_name(value):
    if not value:
        return None
    return PythonQualifiedName(value).package_name


def split_package_name(value):
    return package_name(value)


def base_name(value):
    if not value:
        return None
    return PythonQualifiedName(value).local_name


def split_base_name(value):
    return base_name(value)


def split_package_base_name(value):
    if "." not in value:
        return None, value
    return value.rsplit(".", maxsplit=1)


def qualified_name(package_name: str, class_name: str):
    if "." not in class_name:
        return f"{package_name}.{class_name}"
    if class_name.startswith("."):
        # package_name: aaa.bbb, class name .C => aaa.bbb.C
        # package_name: aaa.bbb, class name ..C => aaa.C
        # package_name: aaa.bbb, class name ...C => C
        class_name = class_name[1:]
        package_path = package_name.split(".")
        while class_name.startswith("."):
            if package_path:
                package_path = package_path[:-1]
            class_name = class_name[1:]
        if package_path:
            class_name = f"{'.'.join(package_path)}.{class_name}"
    return class_name


def convert_config_to_qualified_name(
    config_section, module_field="module", name_field="class"
):
    module = config_section[module_field]
    class_name = config_section[name_field]
    qualified = qualified_name(module, class_name)
    if qualified != class_name:
        config_section[name_field] = qualified


def module_to_path(module):
    return os.path.join(*module.split("."))


def parent_module(module):
    return ".".join(module.split(".")[:-1])


class PythonQualifiedName:
    """
    Helper class to parse python name in the format a.b.C or a.b.c{alias}.
    Can be used anywhere python class can appear, for example:

    base-classes: [invenio.records.Record{InvenioRecord}]
    self.qualified_name = "invenio.records.Record"
    self.local_name = InvenioRecord

    base-classes: [invenio.records.Record]
    self.qualified_name = "invenio.records.Record"
    self.local_name = Record
    """

    def __init__(self, name):
        name = name.strip()
        match = re.match(
            r"^([a-zA-Z_.][a-zA-Z0-9_.]*)({([a-zA-Z_][a-zA-Z0-9_.]*)})?$", name
        )
        if not match:
            raise ValueError(f'Not a python qualified name "{name}"')
        groups = match.groups()
        self.qualified_name = groups[0]
        if groups[2]:
            self.local_name = groups[2]
        else:
            self.local_name = self.qualified_name.rsplit(".", maxsplit=1)[-1]

    @property
    def aliased(self):
        return self.qualified_name.rsplit(".", maxsplit=1)[-1] != self.local_name

    @property
    def package_name(self):
        return self.qualified_name.rsplit(".", maxsplit=1)[0]

    def __str__(self):
        return f"{self.qualified_name}{'{'}{self.local_name}{'}'}"

    @property
    def imports(self):
        if "." not in self.qualified_name:
            return []

        return [
            Import(
                import_path=self.qualified_name,
                alias=self.local_name.split(".")[0] if self.aliased else None,
            )
        ]



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordPermissionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_permissions"
    section = "permissions"
    template = "permissions"



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .record import RecordModelComponent
from .utils import set_default


class RecordDumperClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Generate the dumper class"})
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extensions = ma.fields.List(
        ma.fields.Str(), metadata={"doc": "List of dumper extensions"}
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be copied to the bottom of the dumper file"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class RecordDumperModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, RecordModelComponent]

    class ModelSchema(ma.Schema):
        record_dumper = ma.fields.Nested(
            RecordDumperClassSchema,
            attribute="record-dumper",
            data_key="record-dumper",
            metadata={"doc": "Settings for record dumper"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        record_module = parent_module(datatype.definition["record"]["module"])
        prefix = datatype.definition["module"]["prefix"]

        dumper = set_default(datatype, "record-dumper", {})
        dumper.setdefault("generate", True)

        dumper_module = dumper.setdefault("module", f"{record_module}.dumpers.dumper")
        dumper.setdefault("class", f"{dumper_module}.{prefix}Dumper")
        dumper.setdefault(
            "base-classes", ["oarepo_runtime.records.dumpers.SearchDumper"]
        )
        dumper.setdefault("extra-code", "")
        dumper.setdefault("extensions", [])
        dumper.setdefault("imports", [])



import marshmallow as ma


class PluginEnableDisableField(ma.fields.List):
    def deserialize(self, value, attr, data, **kwargs):
        if value == "__all__":
            return value
        return super().deserialize(value, attr, data, **kwargs)


class PluginSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    include = ma.fields.List(ma.fields.String())
    enable = PluginEnableDisableField(ma.fields.String())
    disable = PluginEnableDisableField(ma.fields.String())


class PluginsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    packages = ma.fields.List(ma.fields.String())
    output = ma.fields.Nested(PluginSchema)
    builder = ma.fields.Nested(PluginSchema)



from typing import Dict

from oarepo_model_builder.datatypes import DataType

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordSearchFacetsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_facets"
    section = "facets"
    template = "record-facets"

    def build_node(self, node: DataType):
        # everything is done in finish
        pass

    def finish(self, **extra_kwargs):
        facets = get_distinct_facets(self.current_model)
        package = self.current_model.definition["facets"]["module"]

        imports = []
        for f in facets:
            imports.extend(f.imports)

        return super().finish(
            current_package_name=package,
            facets=facets,
            facet_imports=imports,
            **extra_kwargs,
        )


def get_distinct_facets(current_model):
    facets_dict: Dict[str, FacetDefinition] = {}
    for node in current_model.deep_iter():
        facet: FacetDefinition
        for facet in node.section_facets.config["facets"]:
            if facet.path not in facets_dict:
                facets_dict[facet.path] = facet
    return list(facets_dict.values())



import marshmallow.fields as fields

from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.datatypes.containers.object import (
    FieldSchema,
    ObjectPropertiesField,
)
from oarepo_model_builder.validation.extensibility import ExtensibleSchema
from oarepo_model_builder.validation.model_validation import ModelFileSchema


def escape(x):
    return x.replace("<", "&lt;").replace(">", "&gt;")


def generate_doc(outfile):
    complete_model_schema = ExtensibleSchema("model_file", ModelFileSchema)
    schemas = {}
    collect_schema(type(complete_model_schema()), schemas)

    for schema_name, schema in schemas.items():
        print(schema_name)
        for fld_name, fld in sorted(schema["fields"].items()):
            print("    ", fld_name, fld)

    with open(outfile, "w") as f:
        for schema_class, schema in schemas.items():
            print(f"## [{schema_class.__name__}](#{schema_class.__name__})", file=f)
            if schema_class.__doc__:
                print(f"{schema_class.__doc__}", file=f)
            print(file=f)
            print("| property | type | documentation |", file=f)
            print("| --- | --- | --- |", file=f)
            for fld_name, fld in sorted(schema["fields"].items()):
                if "ref" in fld:
                    fld_col = f'[{fld["ref"].__name__}](#{fld["ref"].__name__})'
                else:
                    fld_col = fld["type"]
                if "array" in fld:
                    fld_col = f"Array of {fld_col}"
                print(f'| {fld_name} | {fld_col} | {fld.get("doc") } |', file=f)


def collect_schema(schema, schemas):
    if schema in schemas:
        return
    flds = {}
    schemas[schema] = {"fields": flds, "schema": schema}
    if schema is FieldSchema:
        for datatype_name, datatype_class in datatypes.datatype_map.items():
            validator = type(datatype_class.validator())
            collect_schema(validator, schemas)
            flds[f"_{datatype_name}_"] = {
                "type": "Nested",
                "ref": validator,
                "doc": "",
            }
    else:
        for fld_name, fld in schema._declared_fields.items():
            flds[fld.attribute or fld_name] = {
                **collect_field(fld_name, fld, schemas),
                "doc": escape(fld.metadata.get("doc") or ""),
            }


def collect_field(fld_name, fld, schemas):
    if isinstance(fld, fields.Nested):
        subschema = type(fld.schema)
        collect_schema(subschema, schemas)
        return {"type": type(fld).__name__, "ref": subschema}
    elif isinstance(fld, fields.List):
        return {**collect_field(fld_name, fld.inner, schemas), "array": True}
    elif isinstance(fld, ObjectPropertiesField):
        if ObjectPropertiesField not in schemas:
            schemas[ObjectPropertiesField] = {
                "fields": {
                    "*": collect_field(fld_name, fld.value_field, schemas),
                },
                "schema": ObjectPropertiesField,
            }
        return {"type": type(fld).__name__, "ref": ObjectPropertiesField}
    else:
        return {"doc": fld.metadata.get("doc") or fld_name, "type": type(fld).__name__}


if __name__ == "__main__":
    generate_doc("SCHEMA.md")



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioExtResourceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_ext_resource"
    section = "ext"
    template = "ext-resource"

    def finish(self, **extra_kwargs):
        ext = self.current_model.section_ext_resource.config
        super().finish(ext=ext, **extra_kwargs)



import re

import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .record import RecordModelComponent
from .utils import set_default


class PIDSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Set to True (default) provider class"})
    pid_type = ma.fields.String(
        attribute="type",
        data_key="type",
        metadata={"doc": "PID type, generated from model name if not passed"},
    )
    provider_class = ma.fields.String(
        attribute="provider-class",
        data_key="provider-class",
        metadata={"doc": "Fully qualified name of the provider class"},
    )
    provider_base_classes = ma.fields.List(
        ma.fields.String(),
        attribute="provider-base-classes",
        data_key="provider-base-classes",
        metadata={"doc": "List of provider base classes"},
    )
    field_class = ma.fields.String(
        attribute="field-class",
        data_key="field-class",
        metadata={"doc": "Field class, PIDField is used if not passed in"},
    )
    context_class = ma.fields.String(
        attribute="context-class",
        data_key="context-class",
        metadata={"doc": "Context class, PIDFieldContext is used if not passed"},
    )

    field_args = ma.fields.List(
        ma.fields.String(),
        attribute="field-args",
        data_key="field-args",
        metadata={"doc": "Field arguments, default is create=True"},
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "A list of python imports"}
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the pid provider will be placed"}
    )
    skip = ma.fields.Boolean()
    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be copied below the permission class"},
    )


class PIDModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, RecordModelComponent]

    class ModelSchema(ma.Schema):
        pid = ma.fields.Nested(PIDSchema, metadata={"doc": "PID settings"})

    def before_model_prepare(self, datatype, *, context, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        record_module = parent_module(datatype.definition["record"]["module"])

        pid = set_default(datatype, "pid", {})
        pid.setdefault("generate", True)
        pid.setdefault("type", process_pid_type(datatype.definition["model-name"]))

        pid_module = pid.setdefault(
            "module",
            f"{record_module}.api",
        )
        pid.setdefault(
            "provider-class",
            f"{pid_module}.{prefix}IdProvider",
        )
        pid.setdefault(
            "provider-base-classes",
            ["invenio_pidstore.providers.recordid_v2.RecordIdProviderV2"],
        )

        pid.setdefault(
            "field-class", "invenio_records_resources.records.systemfields.pid.PIDField"
        )
        pid.setdefault(
            "context-class",
            "invenio_records_resources.records.systemfields.pid.PIDFieldContext",
        )
        pid.setdefault("field-args", ["create=True"])
        pid.setdefault(
            "imports",
            [],
        )
        pid.setdefault("extra-code", "")


def process_pid_type(pid_base):
    pid_base = re.sub(r"[\s_-]", "", pid_base).lower()
    if len(pid_base) > 6:
        pid_base = re.sub(r"[AEIOU]", "", pid_base, flags=re.IGNORECASE)
    if len(pid_base) > 6:
        pid_base = pid_base[:3] + pid_base[len(pid_base) - 3 :]
    return pid_base



from pathlib import Path

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.builders.utils import ensure_parent_modules
from oarepo_model_builder.utils.jinja import package_name


class PythonBuilder(OutputBuilder):
    def module_to_path(self, module_name):
        mod = module_name.split(".")
        mod[-1] += ".py"
        return Path(*mod)

    def create_parent_modules(self, python_path: Path):
        ensure_parent_modules(
            self.builder, python_path, max_depth=len(python_path.parts)
        )

    def class_to_path(self, class_name):
        return self.module_to_path(package_name(class_name))

    def build_node(self, datatype):
        "intended to be overridden in children"



from pathlib import Path

from oarepo_model_builder.builders import OutputBuilder

from ..utils.dict import dict_get
from .utils import ensure_parent_modules


class JSONBaseBuilder(OutputBuilder):
    output_file_name: str
    output_file_type: str
    parent_module_root_name = None
    create_parent_packages = False

    def begin(self, current_model, schema):
        super().begin(current_model, schema)
        output_name = dict_get(self.current_model.definition, self.output_file_name)
        self.output = self.builder.get_output(self.output_file_type, output_name)

    def finish(self):
        super().finish()

        if self.create_parent_packages and self.output.modified:
            output_name = dict_get(self.current_model.definition, self.output_file_name)
            package_path = Path(output_name).parent
            ensure_parent_modules(
                self.builder,
                self.builder.output_dir.joinpath(package_path / "__init__.py"),
                max_depth=len(package_path.parts),
            )



from .datatypes import DataType


class BaseDateDataType(DataType):
    marshmallow = {"field-class": "marshmallow.fields{ma_fields.String}"}

    facets = {
        "facet-class": "oarepo_runtime.facets.date.DateTimeFacet",
    }


class DateDataType(BaseDateDataType):
    model_type = "date"
    ui = {
        "marshmallow": {
            "field-class": "oarepo_runtime.services.schema.ui.LocalizedDate",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.String}",
        "validators": [
            "{{oarepo_runtime.services.schema.validation.validate_date}}('%Y-%m-%d')"
        ],
    }
    mapping = {"type": "date", "format": "basic_date||strict_date"}
    json_schema = {"type": "string", "format": "date"}


class TimeDataType(BaseDateDataType):
    model_type = "time"

    ui = {
        "marshmallow": {
            "field-class": "oarepo_runtime.services.schema.ui.LocalizedTime",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.String}",
        "validators": [
            "{{oarepo_runtime.services.schema.validation.validate_date}}('%H:%M:%S')"
        ],
    }
    mapping = {
        "type": "date",
        "format": "strict_time||strict_time_no_millis||basic_time||basic_time_no_millis||hour_minute_second||hour||hour_minute",
    }
    json_schema = {"type": "string", "format": "time"}


class DateTimeDataType(BaseDateDataType):
    model_type = "datetime"

    ui = {
        "marshmallow": {
            "field-class": "oarepo_runtime.services.schema.ui.LocalizedDateTime",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.String}",
        "validators": [
            "{{oarepo_runtime.services.schema.validation.validate_datetime}}"
        ],
    }
    mapping = {
        "type": "date",
        "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date||strict_date_hour_minute_second||strict_date_hour_minute_second_fraction",
    }
    json_schema = {"type": "string", "format": "date-time"}


class EDTFDataType(BaseDateDataType):
    model_type = "edtf"

    ui = {
        "marshmallow": {
            "field-class": "oarepo_runtime.services.schema.ui.LocalizedEDTF",
        }
    }
    marshmallow = {
        "field-class": "marshmallow_utils.fields.TrimmedString",
        "validators": [
            "{{oarepo_runtime.services.schema.validation.CachedMultilayerEDTFValidator}}(types=({{edtf.Date{EDTFDate} }},))"
        ],
    }
    mapping = {
        "type": "date",
        "format": "strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy",
    }
    json_schema = {"type": "string", "format": "date-time"}


class EDTFIntervalType(BaseDateDataType):
    model_type = "edtf-interval"

    ui = {
        "marshmallow": {
            "field-class": "oarepo_runtime.services.schema.ui.LocalizedEDTFInterval",
        }
    }
    marshmallow = {
        "field-class": "marshmallow_utils.fields.TrimmedString",
        "validators": [
            "{{oarepo_runtime.services.schema.validation.CachedMultilayerEDTFValidator}}(types=({{edtf.Interval{EDTFInterval} }},))"
        ],
    }
    mapping = {
        "type": "date_range",
        "format": "strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy",
    }
    facets = {
        "searchable": False,  # it seems that facet on edtf is not supported in opensearch
    }
    json_schema = {"type": "string", "format": "date-time"}



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.validation.utils import StrictSchema

from .defaults import DefaultsModelComponent
from .utils import set_default


class ProxySchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    module = ma.fields.Str()
    generate = ma.fields.Boolean()


class ProxyModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(StrictSchema):
        proxy = ma.fields.Nested(
            ProxySchema, metadata={"doc": "Settings for service proxy"}
        )

    def before_model_prepare(self, datatype, **kwargs):
        top_module = datatype.definition["module"]["qualified"]
        proxy = set_default(datatype, "proxy", {})
        proxy.setdefault("module", f"{top_module}.proxies")
        proxy.setdefault("generate", True)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordPIDProviderBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_pid_provider"
    section = "pid"
    template = "pid-provider"



from io import BytesIO, StringIO
from pathlib import Path
from typing import Dict


class AbstractFileSystem:
    def open(self, path, *args, **kwargs):
        raise NotImplementedError()

    def exists(self, path):
        raise NotImplementedError()

    def mkdir(self, path):
        raise NotImplementedError()

    def make_executable(self, path):
        # does nothing on abstract system as it does not have any notion of executable files
        pass


class FileSystem(AbstractFileSystem):
    def __init__(self):
        self.opened_files = set()
        self.overwrite = False

    def open(self, path, mode: str = "r", **kwargs):
        if self.overwrite:
            if "r" in mode and path not in self.opened_files:
                raise FileNotFoundError(f"File {path} not found")
            if "w" in mode or "a" in mode:
                self.opened_files.add(path)
        return open(path, mode=mode, **kwargs)

    def exists(self, path):
        # if overwrite and the file has not been saved, return False
        if self.overwrite:
            return path in self.opened_files
        return Path(path).exists()

    def mkdir(self, path):
        Path(path).mkdir(exist_ok=True, parents=True)

    def make_executable(self, path):
        Path(path).chmod(0o777)


class InMemoryFileSystem(AbstractFileSystem):
    def __init__(self):
        self.files: Dict[str, StringIO] = {}

    def open(self, path: str, mode: str = "r"):
        path = Path(path).absolute()
        if "r" in mode:
            if path not in self.files:
                raise FileNotFoundError(
                    f"File {path} not found. Known files {[f for f in self.files]}"
                )
            if "b" in mode:
                return BytesIO(self.files[path].getvalue())
            return StringIO(self.files[path].getvalue())
        self.files[path] = BytesIO() if "b" in mode else StringIO()
        self.files[path].close = lambda: None
        return self.files[path]

    def exists(self, path):
        path = Path(path).absolute()
        return path in self.files

    def mkdir(self, path):
        # noop on in-memory system
        pass

    def read(self, path):
        with self.open(path) as f:
            return f.read()

    def snapshot(self):
        ret = {}
        for fname, io in self.files.items():
            ret[fname] = io.getvalue()
        return ret
